package trailwalker.track.data;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.PriorityQueue;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.NameValuePair;
import org.apache.http.StatusLine;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.HttpClient;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.message.BasicNameValuePair;

import trailwalker.track.eventHandling.LocationInformationManagerChangeStatusEvent;
import trailwalker.track.eventHandling.LocationInformationManagerChangeStatusEventListener;
import trailwalker.track.ui.data.ApplicationSettingsInformation;

import android.os.AsyncTask;
import android.util.Log;

/**
 * Classe responsable d'emmagatzemar i enviar les posicions rebudes
 * @author Albert Jord Vinardell
 * @version Desembre 2012
 * */
public class LocationInformationManager {
	
	//*******************************************************************************
    //Variables
    //*******************************************************************************
	private PriorityQueue<LocationInformation> locations = new PriorityQueue<LocationInformation>();	
	private int numberOfLocationInformationSent = 0;
	private List<LocationInformationManagerChangeStatusEventListener> _listeners = new ArrayList<LocationInformationManagerChangeStatusEventListener>();
	String rest_URL = "http://napols.uoc.es/arcgis/rest/services/TrailWalker/FeatureServer/0/addFeatures";
	
    //*******************************************************************************
    //Constructor
    //*******************************************************************************
	/**
	 * Constructor
	 * @param settings instncia de {@link ApplicationSettingsInformation} amb la configuraci d'aplicaci
	 * */
	public LocationInformationManager(ApplicationSettingsInformation settings){
		this.useSettings(settings);
	}
	
    //*******************************************************************************
    //Classes Inline
    //*******************************************************************************
	/**
	 * Classe Inline per gestionar amb un fil separat l'enviament de dades al servidor GIS
	 * */
	private class SendStoredLocationsToServer extends AsyncTask<String, Object, Object>{
		@Override
        protected String doInBackground(String... urls) {
        	sendStoredLocationsToServer(urls);
        	return null;
        }
    }
	
    //*******************************************************************************
    //Funcions
    //*******************************************************************************
	/**
	 * Funci per guardar una referncia a un "listener" d'event
	 * @param listener instncia de {@link LocationInformationManagerChangeStatusEventListener} 
	 * */
	public synchronized void addEventListener(LocationInformationManagerChangeStatusEventListener listener) {
		_listeners.add(listener);
	}
	
	/**
	 * Funci per eliminar una referncia a un "listener" d'event
	 * @param listener instncia de {@link LocationInformationManagerChangeStatusEventListener}
	 * */
	public synchronized void removeEventListener(LocationInformationManagerChangeStatusEventListener listener){
		_listeners.remove(listener);
	}
	
	/**
	 * Funci per aplicar i utilitzar els valors de configuraci d'aplicaci
	 * @param settings instncia de {@link ApplicationSettingsInformation} que cont els valors de configuraci
	 * */
	public void useSettings(ApplicationSettingsInformation settings){
		this.rest_URL = settings.getRestUrl();
	}
	
	/**
	 * Funci que torna el nombre de locatlitzacions enviades al servidor GIS
	 * @return nombre de localitzacions enviades al servidor GIS
	 * */
	public int getNumberOfLocationInformationSent(){
		return numberOfLocationInformationSent;
	}
	
	/**
	 * Funci que torna el nombre de localitzacions emmagatzemades (i per tant no enviades)
	 * @return nombre de localitzacions emmagatzemades i  no enviades
	 * */
	public int getNumberOfLocationInformationStored(){
		return locations.size();
	}
	
	/**
	 * Funci que emmagatzema una nova localitzaci i envia totes les localitzacions emmagatzemades
	 * */
	public void storeAndSendLocationToServer(LocationInformation loc){
		Log.d("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: storeAndSendLocationToServer()");
		locations.add(loc);
		statusChanged();
		Log.d("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: storeAndSendLocationToServer(): locations.size() --> " + locations.size());
		new SendStoredLocationsToServer().execute(rest_URL);
	}
	
	/**
	 * Envia totes les localitzacions emmagatzemades i, si l'enviament s correcte, les elimina de memria
	 * @param urls String[] que cont la URL del servidor GIS al que s'ha d'enviar la informaci
	 * */
    public void sendStoredLocationsToServer(String[] urls){
    	Log.d("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: sendInformationToServer()");
    	boolean existsConnection = true;
    	
    	while(!locations.isEmpty()&&existsConnection){
    		LocationInformation li = locations.poll();
    		existsConnection=this.sendLocationInformationToServer(li, urls[0]);
    		if(!existsConnection){
    			locations.add(li);
    		}
    		else {
    			numberOfLocationInformationSent++;
    			statusChanged();
    		}
    	}
    }
    
    /**
     * Funci que realitza una petici HTTP contra una URL passant-li la informaci continguda a "li"
     * @param li instncia de {@link LocationInformation} que cont la informaci a enviar al servidor GIS
     * @param serverURL {@link String} URL del servei REST del servidor GIS
     * @return torna true si l'enviament ha sigut exits, i false si no ho ha sigut
     * */
    private boolean sendLocationInformationToServer(LocationInformation li, String serverURL){
    	Log.d("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: sendLocationInformationToServer()");
    	boolean sent;
    	HttpClient httpClient = new DefaultHttpClient();
    	HttpPost httppost = new HttpPost(serverURL);
    	List<NameValuePair> nameValuePairs = new ArrayList<NameValuePair>(1);
    	nameValuePairs.add(new BasicNameValuePair("f","pjson"));
    	nameValuePairs.add(new BasicNameValuePair("features", "[" + li.toJSONObject().toString() + "]"));
    	Log.d("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: sendLocationInformationToServer(): JSON_request:"+li.toJSONObject().toString()); 
    	try {
			httppost.setEntity(new UrlEncodedFormEntity(nameValuePairs));
			HttpResponse response = httpClient.execute(httppost);
			
			StatusLine sl = response.getStatusLine();
			if(sl.getStatusCode()!=200)return false;
			
			HttpEntity entity = response.getEntity();
			if(entity==null)return false;
			
			InputStream instream = entity.getContent();
			String result = LocationInformationManager.convertStreamToString(instream);
			instream.close();
			
			if(!result.contains("\"success\": true"))return false;
			
			Log.d("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: sendLocationInformationToServer(): JSON_response:"+result); 
			
			sent = true;
			Log.d("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: sendLocationInformationToServer(): location sent to server --> \nLOCATION:" + li.toJSONObject() + "\nServerURL: " + serverURL);
		} catch (UnsupportedEncodingException e) {
			Log.e("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: sendLocationInformationToServer(): UnsupportedEncodingException --> " + e.getMessage());
			sent = false;
		} catch (ClientProtocolException e) {
			Log.e("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: sendLocationInformationToServer(): ClientProtocolException --> " + e.getMessage());
			sent = false;
		} catch (IOException e) {
			Log.e("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: sendLocationInformationToServer(): IOException --> " + e.getMessage());
			sent = false;
		} catch (Exception e){
			Log.e("TrailWalkerTrack", "TrailWalkerTrack: LocationInformationManager: sendLocationInformationToServer(): Exception --> " + e.getMessage());
			sent = false;
		}
		return sent;
    }

    
    /**
     * Funci per enviar un event {@link LocationInformationManagerChangeStatusEvent} als listeners
     * */
	private synchronized void statusChanged() {
		LocationInformationManagerChangeStatusEvent event = new LocationInformationManagerChangeStatusEvent(this);
		Iterator<LocationInformationManagerChangeStatusEventListener> i = _listeners.iterator();
		while(i.hasNext())  {
			((LocationInformationManagerChangeStatusEventListener) i.next()).handleLocationInformationManagerChangeStatusEvent(event);
		}
	}
	
	/**
	 * Funci per convertirun {@link InputStream} en un {@link String}
	 * @param is instncia de {@link InputStream}
	 * @return {@link String} amb la dada convertida
	 * */
	private static String convertStreamToString(InputStream is) {
		BufferedReader reader = new BufferedReader(new InputStreamReader(is));
		StringBuilder sb = new StringBuilder();
		 
		String line = null;
		try {
			while ((line = reader.readLine()) != null) {
			sb.append(line + "\n");
			}
		} 
		catch (IOException e) {
			e.printStackTrace();
		} 
		finally {
			try {
					is.close();
			} catch (IOException e) {
				e.printStackTrace();
				}
		}
		return sb.toString();
	}
	
}
