package trailwalker.track.location;

import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import trailwalker.track.eventHandling.TWLocationManagerEvent;
import trailwalker.track.eventHandling.TWLocationManagerEventListener;
import trailwalker.track.service.TWService;
import trailwalker.track.ui.data.ApplicationSettingsInformation;

import android.content.Context;
import android.location.Location;
import android.location.LocationListener;
import android.location.LocationManager;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.util.Log;

/**
 * Classe que gestiona la captura de posicions
 * @author Albert Jord Vinardell
 * @version Desembre 2012
 * */
public class TWLocationManager {
	
    //*******************************************************************************
    //Constants
    //*******************************************************************************
	final static String TAG = "TrailWalkerTrack: TWLocationManager";
	
    //*******************************************************************************
    //Variables
    //*******************************************************************************

	long minTimeBetweenNetLocUpdates = 0;
	long minDisBetweenNetLocUpdates = 0;
	long minTimeBetweenGPSLocUpdates = 0;
	long minDistBetweenGPSLocUpdates = 0;
	long locationExpirationTime = 0;
	long delayBetweenNetLocUpdatesAndCheckLockStatus = 0;
	
	Context context = null;
	LocationManager locationManager = null;
	LocationListener locationListenerPASSIVE = null;
	LocationListener locationListenerNETWORK = null;
	LocationListener locationListenerGPS = null;
	
	Location lastLocation = null;
	
	Handler handler = null;
	TimerTask task = null;
	Timer timer = null;
	
	boolean locating = false;
	boolean locatingPassive = false;
	boolean locatingNetwork = false;
	boolean locatingGPS = false;
	
	boolean useNetwork = false;
	boolean useGPS = false;
	boolean useGPSFirst = false;
	
	private List<TWLocationManagerEventListener> _listeners = new ArrayList<TWLocationManagerEventListener>();
	
    //*******************************************************************************
    //Constructor
    //*******************************************************************************
	/**
	 * Constructor
	 * @param context instncia de {@link Context} de l'aplicaci
	 * @param settings instncia de {@link ApplicationSettingsInformation} d'on s'extreur el temps entre captures, ...
	 * */
	public TWLocationManager(Context context, ApplicationSettingsInformation settings){
		this.context = context;
		useSettings(settings);
	    initializeComponents();
	}
	
    //*******************************************************************************
    //Funcions
    //*******************************************************************************
	/**
	 * Funci per actualitzar els valors de localitzaci en s pels provets
	 * @param settings instncia de {@link ApplicationSettingsInformation} que cont els valors
	 * */
	public void useSettings(ApplicationSettingsInformation settings){
		minTimeBetweenNetLocUpdates = settings.getTime_between_location_updates();
		minDisBetweenNetLocUpdates = settings.getDistance_between_location_updates();
		minTimeBetweenGPSLocUpdates = settings.getTime_between_location_updates();
		minDistBetweenGPSLocUpdates = settings.getDistance_between_location_updates();
		
		locationExpirationTime = minTimeBetweenNetLocUpdates+(5*minTimeBetweenNetLocUpdates/10); //150% del temps minTimeBetweenNetLocUpdates
		delayBetweenNetLocUpdatesAndCheckLockStatus = 5*minTimeBetweenNetLocUpdates/10; //50% del temps minTimeBetweenNetLocUpdates
		
		useNetwork = settings.isUseNetworkLocation();
		useGPS = settings.isUseGPSLocation();
		useGPSFirst = settings.isUseGPSFirst();
		
		if(locating){
			stopLocationUpdates();
			startLocationUpdates();
		}
	}
	
	/**
	 * Funci per iniciar l'escolta d'events de tipus {@link TWLocationManagerEvent}
	 * */
	public synchronized void addEventListener(TWLocationManagerEventListener listener) {
		_listeners.add(listener);
	}

	/**
	 * Funci per aturar l'escolta d'events de tipus {@link TWLocationManagerEvent}
	 * */	
	public synchronized void removeEventListener(TWLocationManagerEventListener listener){
		_listeners.remove(listener);
	}
	
	/**
	 * Funci per aturar la captura de posicions
	 * */
    public void stopLocationUpdates(){
    	Log.d(TAG, TAG + ": stopLocationUpdates()");
    	if(!locating)return;
    	
    	stopPassiveLocationUpdates();
    	stopNetworkLocationUpdates();
    	stopGPSLocationUpdates();
    	
    	if(timer!=null){
        	timer.cancel();
        	timer.purge();
        	timer = null;
    	}
    	
    	locating = false;
    	
    	raiseLocationServiceStopped();
    }
    
	/**
	 * Funci per iniciar la captura de posicions
	 * */
	public void startLocationUpdates(){
    	Log.d(TAG, TAG + ": startLocationUpdates()");
    	if(locating)return;
    	
    	if(!(checkServices(useGPS, useNetwork)))return;
    	
    	//Primer: iniciem l'escolta passiva de locations
    	startPassiveLocationUpdates();
    	
    	if(useGPS&&useNetwork){
        	//Segon: iniciem el cicle de checks per confirmar que cada X temps tenim una location
    		if(useGPSFirst) startGPSLocationUpdates();
    		else startNetworkLocationUpdates();
        	
        	//Per evitar que intenti programar 2 cops la task
    		timer = new Timer();
        	initializeTask();
    		timer.scheduleAtFixedRate(task, delayBetweenNetLocUpdatesAndCheckLockStatus, minTimeBetweenNetLocUpdates);
    	}
    	else{
    		if(useGPS){
            	//Segon: iniciem el cicle de checks per confirmar que cada X temps tenim una location
            	startGPSLocationUpdates();
    		}
    		else{
    			startNetworkLocationUpdates();
    		}
    	}
    	
    	locating = true;
    	
    	raiseLocationServiceStarted();
    }
	
	/**
	 * Validaci de l'estat dels serveis en el dispositiu
	 * @param GPS true si cal validar l'estat del servei GPS
	 * @param network true si cal validar l'estat del servei de xarxes mbils (localitzaci)
	 * */
	private boolean checkServices(boolean GPS, boolean network){
		if(GPS){
	    	if(!checkIsGPSLocationProviderEnabled()){
	    		stopLocationUpdates();
	    		return false;
	    	}
		}
		
		if(network){
	    	if(!checkIsNetworkLocationProviderEnabled()){
	    		stopLocationUpdates();
	    		return false;
	    	}
		}
		return true;
	}
	
	/**
	 * Funci per inicialitzar la tasca de comprovaci peridica d'antiguitat de l'ltima posici capturada
	 * */
    private void initializeTask(){
        task = new TimerTask() {
			@Override
			public void run() {
				handler.post(new Runnable() {
                    public void run() {
                    	taskExecution();
                    }
				});
			}
		};
    }
    
    /**
     * Implementaci de l'execuci de la tasca de comprovaci de l'antiguitat de l'ltima posici capturada
     * */
    public void taskExecution(){
    	Log.d(TAG, TAG + ": taskExecution()");
    	if(isLastLocationTooOld()){
    		//Si l'ltima posici s massa antiga, s'inicia captura pel sistema alternatiu
    		if(useGPSFirst){
    			Log.d(TAG, TAG + ": taskExecution(): isLastLocationTooOld = true --> Starting Single Network Location Request");
    			startNetworkLocationUpdates();    			
    		}
    		else{
    			Log.d(TAG, TAG + ": taskExecution(): isLastLocationTooOld = true --> Starting Single GPS Location Request");
    			startGPSLocationUpdates();
    		}
    	}
    }
    
    /**
     * Funci que torna boole en funci de si l'ltima posici capturada s massa antiga
     * @return true si l'ltima posici s massa antiga, false si no s aix
     * */
    public boolean isLastLocationTooOld(){
    	Log.d(TAG, TAG + ": isLastLocationTooOld()");
    	if(lastLocation==null)return true;
    	Date d = new Date();
    	return((d.getTime()-lastLocation.getTime())>locationExpirationTime);
    }
    
    /**
     * Funci d'inicialitzaci i implementaci de components com els "escoltadors" de posicions
     * */
    private void initializeComponents(){
    	Log.d(TAG, TAG + ": initializeComponents()");
    	handler = new Handler();
    	locationManager = (LocationManager) context.getSystemService(Context.LOCATION_SERVICE);
    	locationListenerPASSIVE = new LocationListener() {
			@Override
			public void onStatusChanged(String provider, int status, Bundle extras) {}
			@Override
			public void onProviderEnabled(String provider) {}
			@Override
			public void onProviderDisabled(String provider) {}
			@Override
			public void onLocationChanged(Location location) {
				//En cas de rebre una nova localitzaci...
				storeLocation(location);
			}
		};
		locationListenerNETWORK = new LocationListener() {
			@Override
			public void onStatusChanged(String provider, int status, Bundle extras) {}
			@Override
			public void onProviderEnabled(String provider) {startNetworkLocationUpdates();}
			@Override
			public void onProviderDisabled(String provider) {
				//Si es detecta que s'ha aturat el servei de xarxes mbils (localitzaci)...
				stopLocationUpdates();
				raiseNetworkNotEnabled();
			}
			@Override
			public void onLocationChanged(Location location) {
				//En cas de rebre una nova localitzaci: Si utilitzo GPS i Network, i s el prioritari, paro el network
				if(useGPS&&useNetwork&&(useGPSFirst)){
					Log.d(TAG, TAG + ": Stopping Single Network Location Request");
					stopNetworkLocationUpdates();
				}
			}
		};
		
		locationListenerGPS = new LocationListener() {
			
			@Override
			public void onStatusChanged(String provider, int status, Bundle extras) {}
			
			@Override
			public void onProviderEnabled(String provider) {}
			
			@Override
			public void onProviderDisabled(String provider) {
				//Si es detecta que s'ha aturat el servei de GPS...
				stopGPSLocationUpdates();
				raiseGPSNotEnabled();
			}
			@Override
			public void onLocationChanged(Location location) {
				//En cas de rebre una nova localitzaci: Si utilitzo GPS i Network, i GPS no s el prioritari, el paro
				if(useGPS&&useNetwork&&(!useGPSFirst)){
					Log.d(TAG, TAG + ": Stopping Single GPS Location Request");
					stopGPSLocationUpdates();
				}
			}
		};
    }
    
    /**
     * Funci per iniciar l'escolta passiva de noves posicions
     * */
    private void startPassiveLocationUpdates(){
    	Log.d(TAG, TAG + ": startPassiveLocationUpdates()");
    	if(locatingPassive) return;
    	locationManager.requestLocationUpdates(LocationManager.PASSIVE_PROVIDER, 0, 0, locationListenerPASSIVE);
    	locatingPassive = true;
    }
    
    /**
     * Funci per iniciar l'escolta de noves posicions per xarxes mbils (localitzaci)
     * */
    private void startNetworkLocationUpdates(){
    	Log.d(TAG, TAG + ": startNetworkLocationUpdates()");
    	checkIsNetworkLocationProviderEnabled();
    	if(locatingNetwork)return;
    	locationManager.requestLocationUpdates(LocationManager.NETWORK_PROVIDER, minTimeBetweenNetLocUpdates, minDisBetweenNetLocUpdates, locationListenerNETWORK);
    	locatingNetwork = true;
    }
    
    /**
     * Funci per iniciar l'escolta de noves posicions per GPS
     * */
    private void startGPSLocationUpdates(){
    	Log.d(TAG, TAG + ": startGPSLocationUpdates()");
    	checkIsGPSLocationProviderEnabled();
    	if(locatingGPS)return;
    	locationManager.requestLocationUpdates(LocationManager.GPS_PROVIDER, minTimeBetweenGPSLocUpdates, minDistBetweenGPSLocUpdates, locationListenerGPS);
    	locatingGPS = true;
    }
    
    /**
     * Funci per aturar l'escolta passiva de noves posicions
     * */
    private void stopPassiveLocationUpdates(){
    	Log.d(TAG, TAG + ": stopPassiveLocationUpdates()");
    	locationManager.removeUpdates(locationListenerPASSIVE);
    	locatingPassive = false;
    }
    
    /**
     * Funci per aturar l'escolta de noves posicions per xarxes mbils (localitzaci)
     * */
    private void stopNetworkLocationUpdates(){
    	Log.d(TAG, TAG + ": stopNetworkLocationUpdates()");
    	locationManager.removeUpdates(locationListenerNETWORK);
    	locatingNetwork = false;
    }

    /**
     * Funci per iniciar l'escolta de noves posicions per GPS
     * */    
    private void stopGPSLocationUpdates(){
    	Log.d(TAG, TAG + ": stopGPSLocationUpdates()");
    	locationManager.removeUpdates(locationListenerGPS);
    	locatingGPS = false;
    }

    /**
     * Funci per gestionar una nova localitzaci trobada.
     * */
    private void storeLocation(Location loc){
    	Log.d(TAG, TAG + ": storeLocation(): " + loc.toString());
    	lastLocation = loc;
    	
    	//Llana event de que s'ha trobat una nova localitzaci
    	raiseNewLocationFound(loc);
    }

    /**
     * Torna cert si les xarxes mbils (localitzaci) estan activades
     * @return cert si estan activades, fals si no ho estan
     * */    
    private boolean isNetworkLocationProviderEnabled(){
    	return locationManager.isProviderEnabled(LocationManager.NETWORK_PROVIDER);
    }

    /**
     * Torna cert si el GPS est activat
     * @return cert si est activat, fals si no ho est
     * */        
    private boolean isGPSLocationProviderEnabled(){
    	return locationManager.isProviderEnabled(LocationManager.GPS_PROVIDER);
    }
    
    /**
     * Valida si les xarxes mbils (localitzaci) estan activades. Llana un event si no ho estan
     * @return cert o fals en funci de si estan activades o no
     * */    
    private boolean checkIsNetworkLocationProviderEnabled(){
    	Log.d(TAG, TAG + ": checkIsNetworkLocationProviderEnabled()");
    	boolean ret = true;
    	if(!isNetworkLocationProviderEnabled()){
    		raiseNetworkNotEnabled();
    		ret = false;
    	}
    	return ret;
    }

    /**
     * Valida si el GPS est activat. Llana un event si no ho est
     * @return cert o fals en funci de si est activat
     * */    
    private boolean checkIsGPSLocationProviderEnabled(){
    	Log.d(TAG, TAG + ": checkIsGPSLocationProviderEnabled()");
    	boolean ret = true;
    	if(!isGPSLocationProviderEnabled()){
    		raiseGPSNotEnabled();
    		ret = false;
    	}
    	return ret;
    }

    /**
     * Llana un event per indicar que el GPS no est activat
     * */    
    private void raiseGPSNotEnabled(){
    	Log.d(TAG, TAG + ": raiseGPSNotEnabled()");
    	raiseEvent(TWService.MSG_GPS_NOT_ENABLED, null);
    }

    /**
     * Llana un event per indicar que les xarxes mbils (localitzaci) no estan activades
     * */    
    private void raiseNetworkNotEnabled(){
    	Log.d(TAG, TAG + ": raiseNetworkNotEnabled()");
    	raiseEvent(TWService.MSG_NETWORK_NOT_ENABLED, null);
    }
    
    /**
     * Llana un event per indicar que s'ha trobat una nova localitzaci
     * */
    private void raiseNewLocationFound(Location location){
    	Log.d(TAG, TAG + ": raiseNewLocationFound()");
    	raiseEvent(TWService.MSG_LOCATION_RECEIVED, location);
    }
    
    /**
     * Llana un event per indicar que s'ha iniciat el procs de captura de posicions
     * */
    private void raiseLocationServiceStarted(){
    	Log.d(TAG, TAG + ": raiseLocationServiceStarted()");
    	raiseEvent(TWService.MSG_START_LOCATION_UPDATES, null);
    }
    
    /**
     * Llana un event per indicar que s'ha aturat el procs de captura de posicions
     * */
    private void raiseLocationServiceStopped(){
    	Log.d(TAG, TAG + ": raiseLocationServiceStopped()");
    	raiseEvent(TWService.MSG_STOP_LOCATION_UPDATES, null);
    }
    
    /**
     * Funci per llanar events
     * @param messageCode codi numric indicant el tipus de missatge. Constants definides a {@link TWService}
     * @param content instncia d'Object que cont la informaci
     * */
	private synchronized void raiseEvent(int messageCode, Object content) {
    	Message message = new Message();
    	message.what = messageCode;
    	message.obj = content;
		TWLocationManagerEvent event = new TWLocationManagerEvent(this, message);
		Iterator<TWLocationManagerEventListener> i = _listeners.iterator();
		while(i.hasNext())  {
			((TWLocationManagerEventListener) i.next()).handleTWLocationManagerEvent(event);
		}
	}
	  
}
